library nui_database;
export 'package:nui_database/models/enums/nui_database_data_type.dart';
export 'package:nui_database/models/enums/nui_database_operator.dart';
export 'package:nui_database/models/enums/nui_database_query_type.dart';
export 'package:nui_database/models/nui_database_entity.dart';
export 'package:nui_database/models/nui_database_entity_field.dart';
export 'package:nui_database/models/nui_database_entity_mapper.dart';
export 'package:nui_database/models/nui_database_model.dart';
export 'package:nui_database/models/nui_database_query.dart';
export 'package:nui_database/models/nui_database_query_combination.dart';
export 'package:nui_database/models/nui_database_query_interf.dart';
export 'nui_database_callbacks.dart';

import 'dart:async';
import 'dart:developer';
import 'package:flutter/cupertino.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_database/models/nui_database_entity_mapper.dart';
import 'package:nui_database/models/nui_database_model.dart';
import 'package:nui_database/models/nui_database_query_combination.dart';
import 'package:nui_database/nui_database_callbacks.dart';
import 'package:sqflite/sqflite.dart';
import 'package:path/path.dart';
import 'models/enums/nui_query_order.dart';
import 'models/nui_database_entity.dart';
import 'models/nui_database_entity_field.dart';

abstract class NUIDatabaseInterface{
  Future<void> saveEntityLastSync(NUIDBEntityMapper entity);
  Future<DateTime> getEntityLastSync(NUIDBEntityMapper entity);
  Future<void> clearEntityLastSync(NUIDBEntityMapper entity);
  Future<List<T>> getList<T>(NUIDBEntityMapper<T> entity, {InnerLevelQuery condition, String groupBy, String orderBy, NUIQueryOrder order: NUIQueryOrder.ASC, int limit, int offset, DBQueryCallback<T> callback});
  Future<T> getOne<T>(NUIDBEntityMapper<T> entity, {InnerLevelQuery condition, String groupBy, String orderBy, NUIQueryOrder order: NUIQueryOrder.ASC, int offset, DBQueryCallback<T> callback});
  Future<bool> save(NUIDBEntity entity, {DBExecuteCallback callback, bool recordDate:false});
  Future<bool> saveList(List<NUIDBEntity> entities, {DBExecuteCallback callback, bool recordDate:false});
  Future<bool> update(NUIDBEntity entity, {InnerLevelQuery condition, DBExecuteCallback callback});
  Future<bool> updatePartial(NUIDBEntityMapper entity, Map<String, dynamic> updates, {InnerLevelQuery condition, DBExecuteCallback callback});
  Future<bool> clear(NUIDBEntityMapper entity, {DBExecuteCallback callback});
  Future<bool> delete(NUIDBEntityMapper entity, {InnerLevelQuery condition, DBExecuteCallback callback});
}

class NUIDatabase implements NUIDatabaseInterface{

  static const String _MODULE = "NUIDatabase";
  NUIDatabaseModel _db;
  bool _initializingDatabase = false;
  Database database;

  NUIDatabase._builder(NUIDatabaseBuilder builder) : _db = builder._dbModel;

  static Map<String, NUIDatabase> _databaseMap = Map();

  String _getEntityPrefLastSyncKey(NUIDBEntityMapper entity){
    return entity.entityName() + "_lastSync";
  }

  String _getDatabaseTrackerKey(NUIDatabaseModel db){
    return db.databaseName() + "_tracker";
  }

  Future<NUIDatabaseTracker> _saveDatabaseTracker(NUIDatabaseModel db, int version) async{
    final key = _getDatabaseTrackerKey(db);
    final tracker = db.toDatabaseTracker();
    tracker.version = version;
    await NUISharedPref.get().save(key, tracker);
    log("${tracker.toString()}");
    return tracker;
  }

  Future<NUIDatabaseTracker> _getDatabaseTracker(NUIDatabaseModel db) async{
    final key = _getDatabaseTrackerKey(db);
    final tracker = await NUISharedPref.get().getObject(key, NUIDatabaseTrackerEntity());
    if(tracker == null){
      //No previous database version tracker found, so create a new one with version as 1
      return NUIDatabaseTracker(databaseName: db.databaseName(), version: 1, entities: [], isFirst: true);
    }
    else{
      //Found previous database version tracker in shared preference
      return tracker;
    }
  }

  Future<int> _newVersionIfChanged(NUIDatabaseModel db) async {
    final dbTracker = await _getDatabaseTracker(db);
    if (dbTracker.isFirst != true && db.structureChanged(dbTracker)) {
      return dbTracker.version + 1;
    }
    return dbTracker.version;
  }

  Future<bool> _autoMigrateDatabase(int oldVersion, int newVersion, Database database, NUIDatabaseModel db) async{
    final dbTracker = await _getDatabaseTracker(db);

    if(db.structureChanged(dbTracker)){
      //Database changes found, proceed to carry out migration
      final changes = db.getChangedEntities(dbTracker);
      for(NUIDBEntityWithAction entity in changes){
        final table = entity.entity;
        if(match(entity.action, NUIDatabaseModel.ACTION_CREATE)){
          logNUI(_MODULE, "[${_db.databaseName()}] New entity ${entity.tableName} found, adding entity into database ...");
          await _NUIDatabaseUtil.createTable(database, entity.tableName, table.dbFields());
        }
        if(match(entity.action, NUIDatabaseModel.ACTION_RECREATE)){
          logNUI(_MODULE, "[${_db.databaseName()}] Entity structure changed for ${entity.tableName}, recreating entity in database ...");
          final oldEntity = dbTracker.getEntityByName(entity.tableName);
          await _NUIDatabaseUtil.recreateTable(db: database, tableName: entity.tableName, entity: table, oldEntity: oldEntity, patch: false);
          clearEntityLastSync(entity.entity);
        }
        if(match(entity.action, NUIDatabaseModel.ACTION_RECREATE_PATCH)){
          logNUI(_MODULE, "[${_db.databaseName()}] Entity structure changed for ${entity.tableName}, recreating with patching entity in database ...");
          final oldEntity = dbTracker.getEntityByName(entity.tableName);
          await _NUIDatabaseUtil.recreateTable(db: database, tableName: entity.tableName, entity: table, oldEntity: oldEntity, patch: false);
          clearEntityLastSync(entity.entity);
        }
        if(match(entity.action, NUIDatabaseModel.ACTION_REMOVE)){
          logNUI(_MODULE, "[${_db.databaseName()}] Entity ${entity.tableName} has been removed, removing entity from database ...");
          await _NUIDatabaseUtil.dropTable(database, entity.tableName);
          clearEntityLastSync(entity.entity);
        }
        if(match(entity.action, NUIDatabaseModel.ACTION_UPDATE)){
          logNUI(_MODULE, "[${_db.databaseName()}] New columns added to ${entity.tableName}, adding columns to entity in database ...");
          final oldEntity = dbTracker.getEntityByName(entity.tableName);
          final newFields = table.getNewFields(oldEntity);
          await _NUIDatabaseUtil.updateTable(db: database, table: table.entityName(), newFields: newFields);
        }
      }
      await _saveDatabaseTracker(db, newVersion);
      return true;
    }
    return false;
  }

  @override
  Future<void> saveEntityLastSync(NUIDBEntityMapper entity) async{
    NUISharedPref.get().save(_getEntityPrefLastSyncKey(entity), DateTime.now().millisecondsSinceEpoch);
  }

  @override
  Future<DateTime> getEntityLastSync(NUIDBEntityMapper entity) async{
    int lastSync = await NUISharedPref.get().getInt(_getEntityPrefLastSyncKey(entity), defValue: -1);
    if(lastSync != null && lastSync != -1){
      return DateTime.fromMillisecondsSinceEpoch(lastSync);
    }
    return null;
  }

  @override
  Future<void> clearEntityLastSync(NUIDBEntityMapper entity) async{
    await NUISharedPref.get().remove(_getEntityPrefLastSyncKey(entity));
  }

  Future<void> _waitForDatabaseInitialization() async{
    while(database == null){
      logNUI(_MODULE, "[${_db.databaseName()}] Database is null, wait ...");
      await Future.delayed(const Duration(milliseconds: 100));
    }
  }

  static NUIDatabase get(String database){
    NUIDatabase db = _databaseMap[database];
    if(db == null){
      throw Exception("[NUIDatabase] You have not initialized NUIDatabase with builder");
    }
    return db;
  }

  Future<Database> _open(NUIDatabaseModel dbModel, {DBReadyCallback readyCallback}) async {
    try{
      final dbPath = join(await getDatabasesPath(), dbModel.databaseName());
      logNUI(_MODULE, "[${_db.databaseName()}] Opening database ${dbModel.databaseName()} ... ");

      final autoDetectedVersion = await _newVersionIfChanged(dbModel);

      this.database = await openDatabase(
        dbPath,
        version: autoDetectedVersion,
        onCreate: (db, version) async {
          logNUI(_MODULE, "[${_db.databaseName()}] Create database for ${dbModel.databaseName()} with initial version $version");
          await _autoMigrateDatabase(autoDetectedVersion, autoDetectedVersion, db, dbModel);
        },
        onUpgrade: (Database db, int oldVersion, int newVersion) async {
          //Run the migration script to update the tables for each migration
          logNUI(_MODULE, "[${_db.databaseName()}] Upgrade database for ${dbModel.databaseName()} from old version $oldVersion to new version $newVersion");
          await _autoMigrateDatabase(oldVersion, newVersion, db, dbModel);
        },
      );
      logNUI(_MODULE, "[${_db.databaseName()}] Database ${dbModel.databaseName()} opened, current version at $autoDetectedVersion");
      if(readyCallback != null) readyCallback(true, "Open database successful");
      return database;
    }on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Open database failed with error : ${e.toString()}");
      return null;
    }
  }

  @override
  Future<List<T>> getList<T>(NUIDBEntityMapper<T> entity, {InnerLevelQuery condition, String groupBy, String orderBy, NUIQueryOrder order: NUIQueryOrder.ASC, int limit, int offset, DBQueryCallback<T> callback}) async{
    try{
      await _waitForDatabaseInitialization();
      String conditionQuery;
      List<dynamic> params;
      if(condition != null){
        params = List();
        NUIDBQueryCombination conditionComb = NUIDBQueryCombination(queries: List());
        conditionComb = condition(conditionComb);
        if(conditionComb.queries.isNotEmpty){
          conditionQuery = conditionComb.query(params);
        }
      }
      List<T> results = await _NUIDatabaseUtil.query<T>(database, _db, entity, where: conditionQuery, whereArgs: params, groupBy: groupBy, orderBy: orderBy, order: order, limit: limit, offset: offset);
      logNUI(_MODULE, "[${_db.databaseName()}] Query data from ${entity.entityName()} successful with ${results.length} results");
      if(callback != null) callback(results, true, "Query data from ${entity.entityName()} successful");
      return results;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Query data from ${entity.entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(null, false, "Query data from ${entity.entityName()} failed");
      return null;
    }
  }

  @override
  Future<T> getOne<T>(NUIDBEntityMapper<T> entity, {InnerLevelQuery condition, String groupBy, String orderBy, NUIQueryOrder order: NUIQueryOrder.ASC, int offset, DBQueryCallback<T> callback}) async{
    try{
      await _waitForDatabaseInitialization();
      String conditionQuery;
      List<dynamic> params;
      if(condition != null){
        params = List();
        NUIDBQueryCombination conditionComb = NUIDBQueryCombination(queries: List());
        conditionComb = condition(conditionComb);
        if(conditionComb.queries.isNotEmpty){
          conditionQuery = conditionComb.query(params);
        }
      }
      List<T> results = await _NUIDatabaseUtil.query<T>(database, _db, entity, where: conditionQuery, whereArgs: params, groupBy: groupBy, orderBy: orderBy, order: order, limit: 1, offset: offset);
      logNUI(_MODULE, "[${_db.databaseName()}] Query data from ${entity.entityName()} successful with ${results.length} results");
      if(callback != null) callback(results, true, "Query data from ${entity.entityName()} successful");
      if(results != null && results.isNotEmpty){
        return results[0];
      }
      return null;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Query data from ${entity.entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(null, false, "Query data from ${entity.entityName()} failed");
      return null;
    }
  }

  @override
  Future<bool> save(NUIDBEntity entity, {DBExecuteCallback callback, bool recordDate:false}) async{
    try{
      await _waitForDatabaseInitialization();
      await _NUIDatabaseUtil.insert(database, _db, entity);
      logNUI(_MODULE, "[${_db.databaseName()}] Save data of ${entity.mapper().entityName()} successful");
      if(recordDate == true) {
        saveEntityLastSync(entity.mapper());
      }
      if(callback != null) callback(true, "Saved data successfully");
      return true;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Save data of ${entity.mapper().entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(false, "Failed to save data");
      return false;
    }
  }

  @override
  Future<bool> saveList(List<NUIDBEntity> entities, {DBExecuteCallback callback, bool recordDate:false}) async{
    try{
      await _waitForDatabaseInitialization();
      await _NUIDatabaseUtil.insertList(database, _db, entities);
      if(entities != null && entities.isNotEmpty && recordDate == true) {
        saveEntityLastSync(entities[0].mapper());
      }
      if(callback != null) callback(true, "Saved data list successfully");
      return true;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Save data list failed with error : ${e.toString()}");
      if(callback != null) callback(false, "Failed to save data list");
      return false;
    }
  }

  @override
  Future<bool> update(NUIDBEntity entity, {InnerLevelQuery condition, DBExecuteCallback callback}) async{
    try{
      await _waitForDatabaseInitialization();
      String conditionQuery;
      List<dynamic> params;
      if(condition != null){
        params = List();
        NUIDBQueryCombination conditionComb = NUIDBQueryCombination(queries: List());
        conditionComb = condition(conditionComb);
        if(conditionComb.queries.isNotEmpty){
          conditionQuery = conditionComb.query(params);
        }
      }
      await _NUIDatabaseUtil.update(database, _db, entity: entity, entityMapper: entity.mapper(), where: conditionQuery, whereArgs: params);
      logNUI(_MODULE, "[${_db.databaseName()}] Update data of ${entity.mapper().entityName()} successful");
      if(callback != null) callback(true, "Updated data successfully");
      return true;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Update data of ${entity.mapper().entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(false, "Failed to update data");
      return false;
    }
  }

  @override
  Future<bool> updatePartial(NUIDBEntityMapper entity, Map<String, dynamic> updates, {InnerLevelQuery condition, DBExecuteCallback callback}) async{
    try{
      await _waitForDatabaseInitialization();
      String conditionQuery;
      List<dynamic> params;
      if(condition != null){
        params = List();
        NUIDBQueryCombination conditionComb = NUIDBQueryCombination(queries: List());
        conditionComb = condition(conditionComb);
        if(conditionComb.queries.isNotEmpty){
          conditionQuery = conditionComb.query(params);
        }
      }
      await _NUIDatabaseUtil.update(database, _db, entityMapper: entity, updates: updates, where: conditionQuery, whereArgs: params);
      logNUI(_MODULE, "[${_db.databaseName()}] Update data of ${entity.entityName()} successful");
      if(callback != null) callback(true, "Updated data successfully");
      return true;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Update data of ${entity.entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(false, "Failed to update data");
      return false;
    }
  }

  @override
  Future<bool> clear(NUIDBEntityMapper entity, {DBExecuteCallback callback}) async{
    try{
      await _waitForDatabaseInitialization();
      await clearEntityLastSync(entity);
      await _NUIDatabaseUtil.delete(database, _db, entity);
      logNUI(_MODULE, "[${_db.databaseName()}] Clear data of ${entity.entityName()} successful");
      if(callback != null) callback(true, "Clear data successfully");
      return true;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Clear data of ${entity.entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(false, "Failed to clear data");
      return false;
    }
  }

  @override
  Future<bool> delete(NUIDBEntityMapper entity, {InnerLevelQuery condition, DBExecuteCallback callback}) async{
    try{
      await _waitForDatabaseInitialization();
      String conditionQuery;
      List<dynamic> params;
      if(condition != null){
        params = List();
        NUIDBQueryCombination conditionComb = NUIDBQueryCombination(queries: List());
        conditionComb = condition(conditionComb);
        if(conditionComb.queries.isNotEmpty){
          conditionQuery = conditionComb.query(params);
        }
      }
      await _NUIDatabaseUtil.delete(database, _db, entity, where: conditionQuery, whereArgs: params);
      logNUI(_MODULE, "[${_db.databaseName()}] Delete data of ${entity.entityName()} successful");
      if(callback != null) callback(true, "Deleted data successfully");
      return true;
    } on Exception catch(e){
      logNUI(_MODULE, "[${_db.databaseName()}] Delete data of ${entity.entityName()} failed with error : ${e.toString()}");
      if(callback != null) callback(false, "Failed to delete data");
      return false;
    }
  }
}

class NUIDatabaseBuilder{
  NUIDatabaseModel _dbModel;
  DBReadyCallback _readyCallback;

  NUIDatabaseBuilder model(NUIDatabaseModel model){
    this._dbModel = model;
    return this;
  }

  NUIDatabaseBuilder readyCallback(DBReadyCallback callback){
    this._readyCallback = callback;
    return this;
  }

  Future<NUIDatabase> build() async {
    if(_dbModel == null) {
      throw Exception("[NUIDatabase] You must specify a NUIDatabaseModel for the database");
    }

    logNUI(NUIDatabase._MODULE, "[${_dbModel.databaseName()}] Initializing for database ${_dbModel.databaseName()}");
    await NUISharedPrefBuilder().build();
    NUIDatabase db = NUIDatabase._builder(this);
    db._initializingDatabase = true;
    await db._open(_dbModel, readyCallback: _readyCallback);
    NUIDatabase._databaseMap[_dbModel.databaseName()] = db;
    db._initializingDatabase = false;

    logNUI(NUIDatabase._MODULE, "[${_dbModel.databaseName()}] Initialization successful for database ${_dbModel.databaseName()}");
    return db;
  }
}

class _NUIDatabaseUtil{

  static Future<Database> open(String database, List<String> initScripts, {List<String> migrationScripts, int version: 1}) async {
    return openDatabase(
      join(await getDatabasesPath(), database),
      onCreate: (db, version) {
        //Run the initial scripts that create the tables
        initScripts.forEach((script) async {
          await db.execute(script);
        });
      },
      onUpgrade: (Database db, int oldVersion, int newVersion) async {
        //Run the migration script to update the tables for each migration
        for (var i = oldVersion - 1; i <= newVersion - 1; i++) {
          if(migrationScripts.length > i) {
            await db.execute(migrationScripts[i]);
          }
        }
      },
      version: version,
    );
  }

  static String createTableQuery(String tableName, List<NUIDBField> fields, {bool isTemp: false}){
    String query = "CREATE${isTemp ? " TEMPORARY" : ""} TABLE $tableName (";
    for(int i = 0; i < fields.length; i++){
      NUIDBField field = fields[i];
      if(i == fields.length - 1){
        query = query + field.getCreateQuery();
      }
      else{
        query = query + field.getCreateQuery() + ", ";
      }
    }
    query = query + ")";
    return query;
  }

  static String createTableQueryTracker(String tableName, List<NUIDBFieldTracker> fields, {bool isTemp: false}){
    String query = "CREATE${isTemp ? " TEMPORARY" : ""} TABLE $tableName (";
    for(int i = 0; i < fields.length; i++){
      NUIDBFieldTracker field = fields[i];
      if(i == fields.length - 1){
        query = query + field.getCreateQuery();
      }
      else{
        query = query + field.getCreateQuery() + ", ";
      }
    }
    query = query + ")";
    return query;
  }

  static String getColumnsQueryTracker(List<NUIDBFieldTracker> fields){
    String query = "";
    for(int i = 0; i < fields.length; i++){
      NUIDBFieldTracker field = fields[i];
      if(i == fields.length - 1){
        query = query + field.name;
      }
      else{
        query = query + field.name + ",";
      }
    }
    return query;
  }

  static Future<void> createTable(Database db, String tableName, List<NUIDBField> fields, {bool isTemp: false}) async {
    String query = createTableQuery(tableName, fields, isTemp: isTemp) +  "; ";
    await db.execute(query);
  }

  static Future<void> dropTable(Database db, String table) async{
    String query = "DROP TABLE $table; ";
    await db.execute(query);
  }

  static Future<void> updateTable({@required Database db, @required String table, @required List<NUIDBField> newFields}) async{
    if(isNullOrEmpty(newFields)){
      return;
    }

    for(NUIDBField act in newFields){
      await db.execute("ALTER TABLE $table ADD " + act.getCreateQuery() + ";");
    }
  }

  static Future<void> recreateTable({@required Database db, @required String tableName, @required NUIDBEntityMapper entity, @required NUIDBEntityTracker oldEntity, bool patch: true}) async{
    String oriTable = entity.entityName();

    final sameFields = entity.getUnchangedFields(oldEntity);
    String backupTable = oriTable + "_backup";
    if(patch == true) {
      await db.execute(createTableQueryTracker(backupTable, oldEntity.fields, isTemp: true) + "; ");
      await db.execute("INSERT INTO $backupTable SELECT ${getColumnsQueryTracker(oldEntity.fields)} FROM $oriTable; ");
    }
    await db.execute("DROP TABLE $oriTable; ");
    await db.execute(createTableQuery(oriTable, entity.dbFields()) + "; ");
    if(patch == true) {
      await db.execute("INSERT INTO $oriTable SELECT ${getColumnsQueryTracker(sameFields)} FROM $backupTable; ");
      await db.execute("DROP TABLE $backupTable; ");
    }
  }

  static Future<void> insert(Database db, NUIDatabaseModel database, NUIDBEntity entity) async{
    final insertJson = entity.toMapQuery();
    logNUI(NUIDatabase._MODULE, "[${database.databaseName()}] Saving entity ${entity.mapper().entityName()} : $insertJson");
    await db.insert(entity.mapper().entityName(), insertJson, conflictAlgorithm: ConflictAlgorithm.replace);
  }

  static Future<void> insertList(Database db, NUIDatabaseModel database, List<NUIDBEntity> entities) async{
    if(entities == null || entities.isEmpty) return;
    final batch = db.batch();
    for(NUIDBEntity data in entities) {
      final insertJson = data.toMapQuery();
      batch.insert(data.mapper().entityName(), insertJson, conflictAlgorithm: ConflictAlgorithm.replace);
    }
    await batch.commit(noResult: true);
  }

  static Future<void> update(Database db, NUIDatabaseModel database, {NUIDBEntity entity, NUIDBEntityMapper entityMapper, Map<String, dynamic> updates, String where, List<dynamic> whereArgs}) async{
    final Database db = await open(database.databaseName(), []);
    await db.update(entity?.mapper()?.entityName() ?? entityMapper?.entityName(), entity?.toMapQuery() ?? updates, where: where, whereArgs: whereArgs);
  }

  static Future<void> delete(Database db, NUIDatabaseModel database, NUIDBEntityMapper entity, {String where, List<dynamic> whereArgs}) async{
    final Database db = await open(database.databaseName(), []);
    await db.delete(entity.entityName(), where: where, whereArgs: whereArgs);
  }

  static Future<List<T>> query<T>(Database db, NUIDatabaseModel database, NUIDBEntityMapper<T> mapper, {String where, List<dynamic> whereArgs, String groupBy, String orderBy, NUIQueryOrder order: NUIQueryOrder.ASC, int limit, int offset}) async{
    final Database db = await open(database.databaseName(), []);
    String queryOrderBy = (orderBy == null || orderBy.isEmpty) ? null : ("$orderBy " + order.value());
    logNUI(NUIDatabase._MODULE, "[${database.databaseName()}] Querying from entity : ${mapper.entityName()}, condition: $where, arguments : $whereArgs");
    final List<Map<String, dynamic>> maps = await db.query(mapper.entityName(), where: where, whereArgs: whereArgs, groupBy: groupBy, orderBy: queryOrderBy, limit: limit, offset: offset);
    logNUI(NUIDatabase._MODULE, "[${database.databaseName()}] Querying from entity : ${mapper.entityName()} returns with ${maps.length} results");
    return List.generate(maps.length, (i) {
      return mapper.toEntity(maps[i]);
    });
  }

}
